import os
import sys
import re
import copy
import logging

from pydantic import validate_call

from .validator import is_truthy


logger = logging.getLogger(__name__)


@validate_call
def deep_merge(dict1: dict, dict2: dict) -> dict:
    """Return a new dictionary that's the result of a deep merge of two dictionaries.
    If there are conflicts, values from `dict2` will overwrite those in `dict1`.

    Args:
        dict1 (dict, required): The base dictionary that will be merged.
        dict2 (dict, required): The dictionary to merge into `dict1`.

    Returns:
        dict: The merged dictionary.
    """

    _merged = copy.deepcopy(dict1)
    for _key, _val in dict2.items():
        if (
            _key in _merged
            and isinstance(_merged[_key], dict)
            and isinstance(_val, dict)
        ):
            _merged[_key] = deep_merge(_merged[_key], _val)
        else:
            _merged[_key] = copy.deepcopy(_val)

    return _merged


@validate_call
def camel_to_snake(val: str) -> str:
    """Convert CamelCase to snake_case.

    Args:
        val (str): CamelCase string to convert.

    Returns:
        str: Converted snake_case string.
    """

    val = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", val)
    val = re.sub("([a-z0-9])([A-Z])", r"\1_\2", val).lower()
    return val


@validate_call
def get_slug_name(file_path: str | None = None) -> str:
    """Slugify the file name from the given file path or the current script's file path.

    Args:
        file_path (str | None, optional): The file path to slugify. If None, uses the current script's file path.
                                            Defaults to None.

    Returns:
        str: The slugified file name.
    """

    if not file_path:
        file_path = sys.argv[0]

    _slug_name = (
        os.path.splitext(os.path.basename(file_path))[0]
        .strip()
        .replace(" ", "-")
        .replace("_", "-")
        .lower()
    )
    return _slug_name


def is_debug_mode() -> bool:
    """Check if the application is running in debug mode based on environment variables.

    Returns:
        bool: True if in debug mode, False otherwise.
    """

    _is_debug = False
    _debug = os.getenv("DEBUG", "").strip().lower()
    if _debug and is_truthy(_debug):
        _is_debug = True

    _env = os.getenv("ENV", "").strip().lower()
    if (_env == "development") and (_debug == ""):
        _is_debug = True

    return _is_debug


__all__ = [
    "deep_merge",
    "camel_to_snake",
    "get_slug_name",
    "is_debug_mode",
]



REQUEST_ID_REGEX = (
    r"\b[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}\b|"
    r"\b[0-9a-fA-F]{32}\b"
)

# Invalid characters:
SPECIAL_CHARS_REGEX = r"[&'\"<>]"
SPECIAL_CHARS_BASE_REGEX = r"[&'\"<>\\\/]"
SPECIAL_CHARS_LOW_REGEX = r"[&'\"<>\\\/`{}|]"
SPECIAL_CHARS_MEDIUM_REGEX = r"[&'\"<>\\\/`{}|()\[\]]"
SPECIAL_CHARS_HIGH_REGEX = r"[&'\"<>\\\/`{}|()\[\]!@#$%^*;:?]"
SPECIAL_CHARS_STRICT_REGEX = r"[&'\"<>\\\/`{}|()\[\]~!@#$%^*_=\-+;:,.?\t\n ]"


__all__ = [
    "REQUEST_ID_REGEX",
    "SPECIAL_CHARS_REGEX",
    "SPECIAL_CHARS_BASE_REGEX",
    "SPECIAL_CHARS_LOW_REGEX",
    "SPECIAL_CHARS_MEDIUM_REGEX",
    "SPECIAL_CHARS_HIGH_REGEX",
    "SPECIAL_CHARS_STRICT_REGEX",
]



import re
from re import Pattern

from pydantic import validate_call

from .constants import (
    REQUEST_ID_REGEX,
    SPECIAL_CHARS_BASE_REGEX,
    SPECIAL_CHARS_LOW_REGEX,
    SPECIAL_CHARS_MEDIUM_REGEX,
    SPECIAL_CHARS_HIGH_REGEX,
    SPECIAL_CHARS_STRICT_REGEX,
)


@validate_call
def is_truthy(val: str | bool | int | float | None) -> bool:
    """Check if the value is truthy.

    Args:
        val (str | bool | int | float | None, required): Value to check.

    Raises:
        ValueError: If `val` argument type is string and value is invalid.

    Returns:
        bool: True if the value is truthy, False otherwise.
    """

    if isinstance(val, str):
        val = val.strip().lower()

        if val in ["0", "false", "f", "no", "n", "off"]:
            return False
        elif val in ["1", "true", "t", "yes", "y", "on"]:
            return True
        else:
            raise ValueError(f"`val` argument value is invalid: '{val}'!")

    return bool(val)


@validate_call
def is_falsy(val: str | bool | int | float | None) -> bool:
    """Check if the value is falsy.

    Args:
        val (str | bool | int | float | None, required): Value to check.

    Returns:
        bool: True if the value is falsy, False otherwise.
    """

    return not is_truthy(val)


@validate_call
def is_request_id(val: str) -> bool:
    """Check if the string is valid request ID.

    Args:
        val (str, required): String to check.

    Returns:
        bool: True if the string is valid request ID, False otherwise.
    """

    _is_valid = bool(re.match(pattern=REQUEST_ID_REGEX, string=val))
    return _is_valid


@validate_call
def is_blacklisted(val: str, blacklist: list[str]) -> bool:
    """Check if the string is blacklisted.

    Args:
        val       (str      , required): String to check.
        blacklist (List[str], required): List of blacklisted strings.

    Returns:
        bool: True if the string is blacklisted, False otherwise.
    """

    for _blacklisted in blacklist:
        if _blacklisted in val:
            return True

    return False


@validate_call
def is_valid(val: str, pattern: Pattern | str) -> bool:
    """Check if the string is valid with given pattern.

    Args:
        val     (str          , required): String to check.
        pattern (Pattern | str, required): Pattern regex to check.

    Returns:
        bool: True if the string is valid with given pattern, False otherwise.
    """

    _is_valid = bool(re.match(pattern=pattern, string=val))
    return _is_valid


@validate_call
def has_special_chars(val: str, mode: str = "LOW") -> bool:
    """Check if the string has special characters.
    Available modes:
        - "BASE" or "HTML": Basic HTML special characters.
        - "LOW": Low-risk special characters.
        - "MEDIUM": Medium-risk special characters.
        - "HIGH", "SCRIPT", or "SQL": High-risk special characters.
        - "STRICT": Strict mode, checks for most special characters.

    Args:
        val  (str, required): String to check.
        mode (str, optional): Check mode. Defaults to "LOW".

    Raises:
        ValueError: If `mode` argument value is invalid.

    Returns:
        bool: True if the string has special characters, False otherwise.
    """

    _has_special_chars = False

    _pattern = r""
    mode = mode.upper().strip()
    if (mode == "BASE") or (mode == "HTML"):
        _pattern = SPECIAL_CHARS_BASE_REGEX
    elif mode == "LOW":
        _pattern = SPECIAL_CHARS_LOW_REGEX
    elif mode == "MEDIUM":
        _pattern = SPECIAL_CHARS_MEDIUM_REGEX
    elif (mode == "HIGH") or (mode == "SCRIPT") or (mode == "SQL"):
        _pattern = SPECIAL_CHARS_HIGH_REGEX
    elif mode == "STRICT":
        _pattern = SPECIAL_CHARS_STRICT_REGEX
    else:
        raise ValueError(f"`mode` argument value '{mode}' is invalid!")

    _has_special_chars = bool(re.search(pattern=_pattern, string=val))
    return _has_special_chars


__all__ = [
    "is_truthy",
    "is_falsy",
    "is_request_id",
    "is_blacklisted",
    "is_valid",
    "has_special_chars",
]



import hashlib

from pydantic import validate_call

from .constants import HashAlgoEnum


@validate_call
def hash_str(
    val: str | bytes, algorithm: HashAlgoEnum | str = HashAlgoEnum.sha256
) -> str:
    """Hash a string using a specified hash algorithm.

    Args:
        val       (str | bytes       , required): The value to be hashed.
        algorithm (HashAlgoEnum | str, optional): The hash algorithm to use. Defaults to `HashAlgoEnum.sha256`.

    Returns:
        str: The hexadecimal representation of the digest.
    """

    if isinstance(val, str):
        val = val.encode("utf-8")

    if isinstance(algorithm, str):
        algorithm = HashAlgoEnum(algorithm.strip().lower())

    _hash = hashlib.new(algorithm.value)
    _hash.update(val)

    _hash_val = _hash.hexdigest()
    return _hash_val


__all__ = [
    "hash_str",
]



from http import HTTPStatus

from pydantic import validate_call


@validate_call
def get_http_status(status_code: int) -> tuple[HTTPStatus, bool]:
    """Get HTTP status code enum from integer value.

    Args:
        status_code (int, required): Status code for HTTP response: [100 <= status_code <= 599].

    Raises:
        ValueError: If status code is not in range [100 <= status_code <= 599].

    Returns:
        Tuple[HTTPStatus, bool]: Tuple of HTTP status code enum and boolean value if status code is known.
    """

    _http_status: HTTPStatus
    _is_known_status = False
    try:
        _http_status = HTTPStatus(status_code)
        _is_known_status = True
    except ValueError:
        if (100 <= status_code) and (status_code < 200):
            status_code = 100
        elif (200 <= status_code) and (status_code < 300):
            status_code = 200
        elif (300 <= status_code) and (status_code < 400):
            status_code = 304
        elif (400 <= status_code) and (status_code < 500):
            status_code = 400
        elif (500 <= status_code) and (status_code < 600):
            status_code = 500
        else:
            raise ValueError(
                f"`status_code` argument value '{status_code}' is invalid, must be in range 100-599!",
            )

        _http_status = HTTPStatus(status_code)

    return (_http_status, _is_known_status)


__all__ = [
    "get_http_status",
]



from pydantic import validate_call
from starlette.datastructures import URL
from fastapi import Request


@validate_call(config={"arbitrary_types_allowed": True})
def get_relative_url(val: Request | URL) -> str:
    """Get relative url only path with query params from request object or URL object.

    Args:
        val (Request | URL, required): Request object or URL object to extract relative url.

    Returns:
        str: Relative url only path with query params.
    """

    if isinstance(val, Request):
        val = val.url

    _relative_url = str(val).replace(f"{val.scheme}://{val.netloc}", "")
    return _relative_url


__all__ = [
    "get_relative_url",
]



import re
import html
from urllib.parse import quote

from pydantic import validate_call, AnyHttpUrl

from .constants import (
    SPECIAL_CHARS_BASE_REGEX,
    SPECIAL_CHARS_LOW_REGEX,
    SPECIAL_CHARS_MEDIUM_REGEX,
    SPECIAL_CHARS_HIGH_REGEX,
    SPECIAL_CHARS_STRICT_REGEX,
)


@validate_call
def escape_html(val: str) -> str:
    """Escape HTML characters.

    Args:
        val (str, required): String to escape.

    Returns:
        str: Escaped string.
    """

    val = val.strip()
    _escaped = html.escape(val)
    return _escaped


@validate_call
def escape_url(val: AnyHttpUrl | str) -> str:
    """Escape URL characters.

    Args:
        val (AnyHttpUrl, required): String to escape.

    Returns:
        str: Escaped string.
    """

    if isinstance(val, str):
        val = AnyHttpUrl(val)

    _escaped = quote(str(val))
    return _escaped


@validate_call
def sanitize_special_chars(val: str, mode: str = "LOW") -> str:
    """Sanitize special characters.
    Available modes:
        - "BASE" or "HTML": Basic HTML special characters.
        - "LOW": Low-risk special characters.
        - "MEDIUM": Medium-risk special characters.
        - "HIGH", "SCRIPT", or "SQL": High-risk special characters.
        - "STRICT": Strict mode, removes most special characters.

    Args:
        val  (str, required): String to sanitize.
        mode (str, optional): Sanitization mode. Defaults to "LOW".

    Raises:
        ValueError: If `mode` argument value is invalid.

    Returns:
        str: Sanitized string.
    """

    _pattern = r""
    mode = mode.upper().strip()
    if (mode == "BASE") or (mode == "HTML"):
        _pattern = SPECIAL_CHARS_BASE_REGEX
    elif mode == "LOW":
        _pattern = SPECIAL_CHARS_LOW_REGEX
    elif mode == "MEDIUM":
        _pattern = SPECIAL_CHARS_MEDIUM_REGEX
    elif (mode == "HIGH") or (mode == "SCRIPT") or (mode == "SQL"):
        _pattern = SPECIAL_CHARS_HIGH_REGEX
    elif mode == "STRICT":
        _pattern = SPECIAL_CHARS_STRICT_REGEX
    else:
        raise ValueError(f"`mode` argument value '{mode}' is invalid!")

    _sanitized = re.sub(pattern=_pattern, repl="", string=val)
    return _sanitized


__all__ = [
    "escape_html",
    "escape_url",
    "sanitize_special_chars",
]



import uuid
import string
import secrets

from pydantic import validate_call

from .dt import now_ts


@validate_call
def gen_unique_id(prefix: str = "") -> str:
    """Generate unique ID. Format: '{prefix}{datetime}_{uuid4}'.

    Args:
        prefix (str, optional): Prefix of ID. Defaults to ''.

    Raises:
        ValueError: If `prefix` length is greater than 32.

    Returns:
        str: Unique ID.
    """

    prefix = prefix.strip()
    if 32 < len(prefix):
        raise ValueError(
            f"`prefix` argument length {len(prefix)} is too long, must be less than or equal to 32!",
        )

    _id = str(f"{prefix}{now_ts()}_{uuid.uuid4().hex}").lower()
    return _id


@validate_call
def gen_random_string(length: int = 16, is_alphanum: bool = True) -> str:
    """Generate secure random string.

    Args:
        length      (int , optional): Length of random string. Defaults to 16.
        is_alphanum (bool, optional): If True, generate only alphanumeric string. Defaults to True.

    Raises:
        ValueError: If `length` is less than 1.

    Returns:
        str: Generated random string.
    """

    if length < 1:
        raise ValueError(
            f"`length` argument value {length} is too small, must be greater than or equal to 1!",
        )

    _base_chars = string.ascii_letters + string.digits
    if not is_alphanum:
        _base_chars += string.punctuation

    _random_str = "".join(secrets.choice(_base_chars) for _i in range(length))
    return _random_str


__all__ = [
    "gen_unique_id",
    "gen_random_string",
]



import time
import logging
from zoneinfo import ZoneInfo
from datetime import datetime, timezone, tzinfo, timedelta

from pydantic import validate_call

from .constants import WarnEnum, TSUnitEnum


logger = logging.getLogger(__name__)


def now_local_dt() -> datetime:
    """Get current datetime in local timezone with tzinfo.

    Returns:
        datetime: Current datetime in local timezone with tzinfo.
    """

    _local_dt = datetime.now().astimezone()
    return _local_dt


def now_utc_dt() -> datetime:
    """Get current datetime in UTC timezone with tzinfo.

    Returns:
        datetime: Current datetime in UTC timezone with tzinfo.
    """

    _utc_dt = datetime.now(tz=timezone.utc)
    return _utc_dt


@validate_call(config={"arbitrary_types_allowed": True})
def now_dt(tz: ZoneInfo | tzinfo | str | None = None) -> datetime:
    """Get current datetime in specified timezone with tzinfo.

    Args:
        tz (ZoneInfo | tzinfo | str | None, optional): Timezone info. Defaults to None (UTC timezone).

    Returns:
        datetime: Current datetime in specified timezone with tzinfo.
    """

    _dt = now_utc_dt()
    if tz:
        _dt = convert_tz(dt=_dt, tz=tz)
    return _dt


@validate_call(config={"arbitrary_types_allowed": True})
def replace_tz(dt: datetime, tz: ZoneInfo | tzinfo | str) -> datetime:
    """Add or replace timezone info to datetime object.

    Args:
        dt (datetime               , required): Datetime object.
        tz (ZoneInfo | tzinfo | str, required): Timezone info.

    Raises:
        ZoneInfoNotFoundError: If `tz` argument value is invalid.

    Returns:
        datetime: Datetime object with timezone info.
    """

    if isinstance(tz, str):
        tz = ZoneInfo(tz)

    dt = dt.replace(tzinfo=tz)
    return dt


@validate_call(config={"arbitrary_types_allowed": True})
def convert_tz(
    dt: datetime,
    tz: ZoneInfo | tzinfo | str,
    warn_mode: WarnEnum | str = WarnEnum.ALWAYS,
) -> datetime:
    """Convert datetime object to another timezone.

    Args:
        dt        (datetime               , required): Datetime object to convert.
        tz        (ZoneInfo | tzinfo | str, required): Timezone info to convert.
        warn_mode (WarnEnum | str         , optional): Warning mode. Defaults to WarnEnum.ALWAYS.

    Raises:
        ValueError: If `dt` argument doesn't have any timezone info and `warn_mode` is set to WarnEnum.ERROR.
        ValueError: If `warn_mode` argument value is invalid.

    Returns:
        datetime: Datetime object which has been converted to another timezone.
    """

    if isinstance(warn_mode, str):
        warn_mode = WarnEnum(warn_mode.strip().upper())

    if not dt.tzinfo:
        _message = "Not found any timezone info in `dt` argument, assuming it's UTC timezone..."
        if warn_mode == WarnEnum.ALWAYS:
            logger.warning(_message)
        elif warn_mode == WarnEnum.DEBUG:
            logger.debug(_message)
        elif warn_mode == WarnEnum.ERROR:
            _message = "Not found any timezone info in `dt` argument!"
            logger.error(_message)
            raise ValueError(_message)

        dt = replace_tz(dt=dt, tz="UTC")

    if isinstance(tz, str):
        tz = ZoneInfo(tz)

    dt = dt.astimezone(tz=tz)
    return dt


@validate_call
def dt_to_iso(
    dt: datetime, sep: str = "T", warn_mode: WarnEnum | str = WarnEnum.IGNORE
) -> str:
    """Convert datetime object to ISO 8601 format.

    Args:
        dt        (datetime      , required): Datetime object.
        sep       (str           , optional): Separator between date and time. Defaults to "T".
        warn_mode (WarnEnum | str, optional): Warning mode. Defaults to WarnEnum.IGNORE.

    Raises:
        ValueError: If `dt` argument doesn't have any timezone info and `warn_mode` is set to WarnEnum.ERROR.
        ValueError: If `sep` argument length is greater than 8.
        ValueError: If `warn_mode` argument value is invalid.

    Returns:
        str: Datetime string in ISO 8601 format.
    """

    sep = sep.strip()
    if 8 < len(sep):
        raise ValueError(
            f"`sep` argument length '{len(sep)}' is too long, must be less than or equal to 8!"
        )

    if isinstance(warn_mode, str):
        warn_mode = WarnEnum(warn_mode.strip().upper())

    if not dt.tzinfo:
        _message = "Not found any timezone info in `dt` argument, assuming it's UTC timezone..."
        if warn_mode == WarnEnum.ALWAYS:
            logger.warning(_message)
        elif warn_mode == WarnEnum.DEBUG:
            logger.debug(_message)
        elif warn_mode == WarnEnum.ERROR:
            _message = "Not found any timezone info in `dt` argument!"
            logger.error(_message)
            raise ValueError(_message)

        dt = replace_tz(dt=dt, tz="UTC")

    _dt_str = dt.isoformat(sep=sep, timespec="milliseconds")
    return _dt_str


@validate_call(config={"arbitrary_types_allowed": True})
def calc_future_dt(
    delta: timedelta | int,
    dt: datetime | None = None,
    tz: ZoneInfo | tzinfo | str | None = None,
) -> datetime:
    """Calculate future datetime by adding delta time to current or specified datetime.

    Args:
        delta (timedelta | int             , required): Delta time to add to current or specified datetime.
        dt    (datetime | None             , optional): Datetime before adding delta time. Defaults to None.
        tz    (ZoneInfo | tzinfo, str, None, optional): Timezone info. Defaults to None.

    Returns:
        datetime: Calculated future datetime.
    """

    if not dt:
        dt = now_utc_dt()

    if tz:
        dt = convert_tz(dt=dt, tz=tz)

    if isinstance(delta, int):
        delta = timedelta(seconds=delta)

    _future_dt = dt + delta
    return _future_dt


@validate_call
def now_ts(unit: TSUnitEnum | str = TSUnitEnum.SECONDS) -> int:
    """Get current timestamp in UTC timezone.

    Args:
        unit (TSUnitEnum | str, optional): Type of timestamp unit. Defaults to TSUnitEnum.SECONDS.

    Raises:
        ValueError: If `unit` argument value is invalid.

    Returns:
        int: Current timestamp.
    """

    _now_ts: int
    if unit == TSUnitEnum.SECONDS:
        _now_ts = int(time.time())
    elif unit == TSUnitEnum.MILLISECONDS:
        _now_ts = int(time.time() * 1000)
    elif unit == TSUnitEnum.MICROSECONDS:
        _now_ts = int(time.time_ns() / 1000)
    elif unit == TSUnitEnum.NANOSECONDS:
        _now_ts = int(time.time_ns())
    else:
        raise ValueError(f"`unit` argument value '{unit}' is invalid!")

    return _now_ts


@validate_call
def dt_to_ts(dt: datetime, unit: TSUnitEnum | str = TSUnitEnum.SECONDS) -> int:
    """Convert datetime to timestamp.

    Args:
        dt   (datetime        , required): Datetime object to convert.
        unit (TSUnitEnum | str, optional): Type of timestamp unit. Defaults to `TSUnitEnum.SECONDS`.

    Raises:
        ValueError: If `unit` argument value is invalid.

    Returns:
        int: Converted timestamp.
    """

    _ts: int
    if unit == TSUnitEnum.SECONDS:
        _ts = int(dt.timestamp())
    elif unit == TSUnitEnum.MILLISECONDS:
        _ts = int(dt.timestamp() * 1000)
    elif unit == TSUnitEnum.MICROSECONDS:
        _ts = int(dt.timestamp() * 1000000)
    elif unit == TSUnitEnum.NANOSECONDS:
        _ts = int(dt.timestamp() * 1000000000)
    else:
        raise ValueError(f"`unit` argument value '{unit}' is invalid!")

    return _ts


__all__ = [
    "now_utc_dt",
    "now_local_dt",
    "now_dt",
    "replace_tz",
    "convert_tz",
    "dt_to_iso",
    "calc_future_dt",
    "now_ts",
    "dt_to_ts",
]



# noqa: E402

import os
import sys
import json
import glob
import errno
import shutil
import hashlib
import logging
import configparser
from pathlib import Path
from typing import Any

_binary_toml = False
if sys.version_info >= (3, 11):
    import tomllib  # type: ignore

    _binary_toml = True
else:
    import toml as tomllib  # type: ignore

import yaml
from pydantic import validate_call

from .._base import deep_merge
from ..constants import WarnEnum, HashAlgoEnum, ConfigFileFormatEnum, MAX_PATH_LENGTH


logger = logging.getLogger(__name__)


@validate_call
def create_dir(create_dir: str, warn_mode: WarnEnum | str = WarnEnum.DEBUG) -> None:
    """Create directory if `create_dir` doesn't exist.

    Args:
        create_dir (str           , required): Create directory path.
        warn_mode  (WarnEnum | str, optional): Warning message mode, for example: 'ERROR', 'ALWAYS', 'DEBUG', 'IGNORE'.
                                                Defaults to 'DEBUG'.

    Raises:
        ValueError: If `create_dir` argument length is out of range.
        OSError   : When warning mode is set to ERROR and directory already exists.
        OSError   : If failed to create directory.
    """

    create_dir = create_dir.strip()
    if (len(create_dir) < 1) or (MAX_PATH_LENGTH < len(create_dir)):
        raise ValueError(
            f"`create_dir` argument length {len(create_dir)} is out of range, "
            f"must be between 1 and {MAX_PATH_LENGTH} characters!"
        )

    if isinstance(warn_mode, str):
        warn_mode = WarnEnum(warn_mode.strip().upper())

    if not os.path.isdir(create_dir):
        try:
            _message = f"Creating '{create_dir}' directory..."
            if warn_mode == WarnEnum.ALWAYS:
                logger.info(_message)
            elif warn_mode == WarnEnum.DEBUG:
                logger.debug(_message)

            os.makedirs(create_dir)
        except OSError as err:
            if (err.errno == errno.EEXIST) and (warn_mode == WarnEnum.DEBUG):
                logger.debug(f"'{create_dir}' directory already exists!")
            else:
                logger.error(f"Failed to create '{create_dir}' directory!")
                raise

        _message = f"Successfully created '{create_dir}' directory."
        if warn_mode == WarnEnum.ALWAYS:
            logger.info(_message)
        elif warn_mode == WarnEnum.DEBUG:
            logger.debug(_message)

    elif warn_mode == WarnEnum.ERROR:
        raise OSError(errno.EEXIST, f"'{create_dir}' directory already exists!")

    return


@validate_call
def remove_dir(remove_dir: str, warn_mode: WarnEnum | str = WarnEnum.DEBUG) -> None:
    """Remove directory if `remove_dir` exists.

    Args:
        remove_dir (str           , required): Remove directory path.
        warn_mode  (WarnEnum | str, optional): Warning message mode, for example: 'ERROR', 'ALWAYS', 'DEBUG', 'IGNORE'.
                                                Defaults to 'DEBUG'.

    Raises:
        ValueError: If `remove_dir` argument length is out of range.
        OSError   : When warning mode is set to ERROR and directory doesn't exist.
        OSError   : If failed to remove directory.
    """

    remove_dir = remove_dir.strip()
    if (len(remove_dir) < 1) or (MAX_PATH_LENGTH < len(remove_dir)):
        raise ValueError(
            f"`remove_dir` argument length {len(remove_dir)} is out of range, "
            f"must be between 1 and {MAX_PATH_LENGTH} characters!"
        )

    if isinstance(warn_mode, str):
        warn_mode = WarnEnum(warn_mode.strip().upper())

    if os.path.isdir(remove_dir):
        try:
            _message = f"Removing '{remove_dir}' directory..."
            if warn_mode == WarnEnum.ALWAYS:
                logger.info(_message)
            elif warn_mode == WarnEnum.DEBUG:
                logger.debug(_message)

            shutil.rmtree(remove_dir)
        except OSError as err:
            if (err.errno == errno.ENOENT) and (warn_mode == WarnEnum.DEBUG):
                logger.debug(f"'{remove_dir}' directory doesn't exist!")
            else:
                logger.error(f"Failed to remove '{remove_dir}' directory!")
                raise

        _message = f"Successfully removed '{remove_dir}' directory."
        if warn_mode == WarnEnum.ALWAYS:
            logger.info(_message)
        elif warn_mode == WarnEnum.DEBUG:
            logger.debug(_message)

    elif warn_mode == WarnEnum.ERROR:
        raise OSError(errno.ENOENT, f"'{remove_dir}' directory doesn't exist!")

    return


@validate_call
def remove_dirs(
    remove_dirs: list[str], warn_mode: WarnEnum | str = WarnEnum.DEBUG
) -> None:
    """Remove directories if `remove_dirs` exist.

    Args:
        remove_dirs (list[str]     , required): Remove directory paths as list.
        warn_mode   (WarnEnum | str, optional): Warning message mode, for example: 'ERROR', 'ALWAYS', 'DEBUG', 'IGNORE'.
                                                Defaults to 'DEBUG'.
    """

    for _remove_dir in remove_dirs:
        remove_dir(remove_dir=_remove_dir, warn_mode=warn_mode)

    return


@validate_call
def remove_file(file_path: str, warn_mode: WarnEnum | str = WarnEnum.DEBUG) -> None:
    """Remove file if `file_path` exists.

    Args:
        file_path (str           , required): Remove file path.
        warn_mode (WarnEnum | str, optional): Warning message mode, for example: 'ERROR', 'ALWAYS', 'DEBUG', 'IGNORE'.
                                                Defaults to 'DEBUG'.

    Raises:
        ValueError: If `file_path` argument length is out of range.
        OSError   : When warning mode is set to ERROR and file doesn't exist.
        OSError   : If failed to remove file.
    """

    file_path = file_path.strip()
    if (len(file_path) < 1) or (MAX_PATH_LENGTH < len(file_path)):
        raise ValueError(
            f"`file_path` argument length {len(file_path)} is out of range, "
            f"must be between 1 and {MAX_PATH_LENGTH} characters!"
        )

    if isinstance(warn_mode, str):
        warn_mode = WarnEnum(warn_mode.strip().upper())

    if os.path.isfile(file_path):
        try:
            _message = f"Removing '{file_path}' file..."
            if warn_mode == WarnEnum.ALWAYS:
                logger.info(_message)
            elif warn_mode == WarnEnum.DEBUG:
                logger.debug(_message)

            os.remove(file_path)
        except OSError as err:
            if (err.errno == errno.ENOENT) and (warn_mode == WarnEnum.DEBUG):
                logger.debug(f"'{file_path}' file doesn't exist!")
            else:
                logger.error(f"Failed to remove '{file_path}' file!")
                raise

        _message = f"Successfully removed '{file_path}' file."
        if warn_mode == WarnEnum.ALWAYS:
            logger.info(_message)
        elif warn_mode == WarnEnum.DEBUG:
            logger.debug(_message)

    elif warn_mode == WarnEnum.ERROR:
        raise OSError(errno.ENOENT, f"'{file_path}' file doesn't exist!")

    return


@validate_call
def remove_files(
    file_paths: list[str], warn_mode: WarnEnum | str = WarnEnum.DEBUG
) -> None:
    """Remove files if `file_paths` exist.

    Args:
        file_paths (list[str]     , required): Remove file paths as list.
        warn_mode  (WarnEnum | str, optional): Warning message mode, for example: 'ERROR', 'ALWAYS', 'DEBUG', 'IGNORE'.
                                                Defaults to 'DEBUG'.
    """

    for _file_path in file_paths:
        remove_file(file_path=_file_path, warn_mode=warn_mode)

    return


@validate_call
def get_file_checksum(
    file_path: str,
    hash_method: HashAlgoEnum = HashAlgoEnum.md5,
    chunk_size: int = 4096,
    warn_mode: WarnEnum | str = WarnEnum.DEBUG,
) -> str | None:
    """Get file checksum.

    Args:
        file_path   (str           , required): Target file path.
        hash_method (HashAlgoEnum  , optional): Hash method. Defaults to `HashAlgoEnum.md5`.
        chunk_size  (int           , optional): Chunk size. Defaults to 4096.
        warn_mode   (WarnEnum | str, optional): Warning message mode, for example: 'ERROR', 'ALWAYS', 'DEBUG', 'IGNORE'.
                                                    Defaults to 'DEBUG'.

    Raises:
        ValueError: If `file_path` argument length is out of range.
        ValueError: If `chunk_size` argument value is invalid.
        OSError   : When warning mode is set to ERROR and file doesn't exist.

    Returns:
        str | None: File checksum or None if file doesn't exist.
    """

    file_path = file_path.strip()
    if (len(file_path) < 1) or (MAX_PATH_LENGTH < len(file_path)):
        raise ValueError(
            f"`file_path` argument length {len(file_path)} is out of range, "
            f"must be between 1 and {MAX_PATH_LENGTH} characters!"
        )

    if chunk_size < 10:
        raise ValueError(
            f"`chunk_size` argument value {chunk_size} is invalid, must be greater than 10!"
        )

    if isinstance(warn_mode, str):
        warn_mode = WarnEnum(warn_mode.strip().upper())

    _file_checksum: str | None = None
    if os.path.isfile(file_path):
        _file_hash = hashlib.new(hash_method.value)
        with open(file_path, "rb") as _file:
            while True:
                _file_chunk = _file.read(chunk_size)
                if not _file_chunk:
                    break
                _file_hash.update(_file_chunk)

        _file_checksum = _file_hash.hexdigest()
    else:
        _message = f"'{file_path}' file doesn't exist!"
        if warn_mode == WarnEnum.ALWAYS:
            logger.warning(_message)
        elif warn_mode == WarnEnum.DEBUG:
            logger.debug(_message)
        elif warn_mode == WarnEnum.ERROR:
            raise OSError(errno.ENOENT, _message)

    return _file_checksum


@validate_call
def read_yaml_file(file_path: str | Path) -> dict[str, Any]:
    """Read YAML file.

    Args:
        file_path (str | Path, required): YAML file path.

    Raises:
        FileNotFoundError: If YAML file is not found.
        Exception        : If failed to read YAML file.

    Returns:
        dict[str, Any]: YAML file data as dictionary.
    """

    _data: dict[str, Any] = {}

    if isinstance(file_path, str):
        file_path = Path(file_path)

    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"Not found '{file_path}' YAML file!")

    try:
        with open(file_path, encoding="utf-8") as _file:
            _data = yaml.safe_load(_file) or {}
    except Exception:
        logger.error(f"Failed to read '{file_path}' YAML file!")
        raise

    return _data


@validate_call
def read_json_file(file_path: str | Path) -> dict[str, Any]:
    """Read JSON file.

    Args:
        file_path (str | Path, required): JSON file path.

    Raises:
        FileNotFoundError: If JSON file is not found.
        Exception        : If failed to read JSON file.

    Returns:
        dict[str, Any]: JSON file data as dictionary.
    """

    _data: dict[str, Any] = {}

    if isinstance(file_path, str):
        file_path = Path(file_path)

    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"Not found '{file_path}' JSON file!")

    try:
        with open(file_path, encoding="utf-8") as _file:
            _data = json.load(_file) or {}
    except Exception:
        logger.error(f"Failed to read '{file_path}' JSON file!")
        raise

    return _data


@validate_call
def read_toml_file(file_path: str | Path) -> dict[str, Any]:
    """Read TOML file.

    Args:
        file_path (str | Path, required): TOML file path.

    Raises:
        FileNotFoundError: If TOML file is not found.
        Exception        : If failed to read TOML file.

    Returns:
        dict[str, Any]: TOML file data as dictionary.
    """

    _data: dict[str, Any] = {}

    if isinstance(file_path, str):
        file_path = Path(file_path)

    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"Not found '{file_path}' TOML file!")

    try:
        if _binary_toml:
            with open(file_path, "rb") as _file:
                _data = tomllib.load(_file) or {}  # type: ignore
        else:
            with open(file_path, encoding="utf-8") as _file:
                _data = tomllib.load(_file) or {}  # type: ignore
    except Exception:
        logger.error(f"Failed to read '{file_path}' TOML file!")
        raise

    return _data


@validate_call
def read_ini_file(file_path: str | Path) -> dict[str, Any]:
    """Read INI config file.

    Args:
        file_path (str | Path, required): INI config file path.

    Raises:
        FileNotFoundError: If INI config file is not found.
        Exception        : If failed to read INI config file.

    Returns:
        dict[str, Any]: INI config file data as dictionary.
    """

    _config: dict[str, Any] = {}

    if isinstance(file_path, str):
        file_path = Path(file_path)

    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"Not found '{file_path}' INI config file!")

    try:
        _config_parser = configparser.ConfigParser()
        _config_parser.read(file_path)
        for _section in _config_parser.sections():
            _config[_section] = dict(_config_parser.items(_section))

    except Exception:
        logger.error(f"Failed to read '{file_path}' INI config file!")
        raise

    return _config


@validate_call
def read_config_file(config_path: str | Path) -> dict[str, Any]:
    """Read config file (YAML, JSON, TOML, INI).

    Args:
        config_path (str | Path, required): Config file path.

    Raises:
        FileNotFoundError: If config file is not found.
        ValueError       : If config file format is not supported.

    Returns:
        dict[str, Any]: Config file data as dictionary.
    """

    _config: dict[str, Any] = {}

    if isinstance(config_path, str):
        config_path = Path(config_path)

    if not os.path.isfile(config_path):
        raise FileNotFoundError(f"Not found '{config_path}' config file!")

    _suffix = config_path.suffix.lower()
    if _suffix in (".yaml", ".yml"):
        _config = read_yaml_file(file_path=config_path)
    elif _suffix == ".json":
        _config = read_json_file(file_path=config_path)
    elif _suffix == ".toml":
        _config = read_toml_file(file_path=config_path)
    elif _suffix in (".ini", ".cfg"):
        _config = read_ini_file(file_path=config_path)
    else:
        raise ValueError(
            f"Unsupported config file format '{config_path.suffix}' for '{config_path}'!"
        )

    return _config


@validate_call
def read_all_configs(
    configs_dir: str | Path | list[str | Path],
    allowed_formats: list[ConfigFileFormatEnum] = [
        ConfigFileFormatEnum.YAML,
        ConfigFileFormatEnum.JSON,
        ConfigFileFormatEnum.TOML,
    ],
) -> dict[str, Any]:
    """Read all config files from directory or directories and merge them.

    Args:
        configs_dir     (str | Path | list[str | Path], required): Configs directory or directories.
        allowed_formats (list[ConfigFileFormatEnum]   , optional): Allowed config file formats to read.
                                                                    Defaults to [YAML, JSON, TOML].

    Returns:
        dict[str, Any]: Dictionary containing all merged config data from all files.
    """

    _config_dict: dict[str, Any] = {}

    if not isinstance(configs_dir, list):
        configs_dir = [configs_dir]

    _file_paths: list[str] = []
    for _config_dir in configs_dir:
        if isinstance(_config_dir, str):
            _config_dir = Path(_config_dir)

        if not os.path.isabs(_config_dir):
            _current_dir = os.getcwd()
            _config_dir = os.path.join(_current_dir, _config_dir)

        if os.path.isdir(_config_dir):
            if ConfigFileFormatEnum.YAML in allowed_formats:
                _file_paths.extend(glob.glob(os.path.join(_config_dir, "*.yaml")))
                _file_paths.extend(glob.glob(os.path.join(_config_dir, "*.yml")))

            if ConfigFileFormatEnum.JSON in allowed_formats:
                _file_paths.extend(glob.glob(os.path.join(_config_dir, "*.json")))

            if ConfigFileFormatEnum.TOML in allowed_formats:
                _file_paths.extend(glob.glob(os.path.join(_config_dir, "*.toml")))

            if ConfigFileFormatEnum.INI in allowed_formats:
                _file_paths.extend(glob.glob(os.path.join(_config_dir, "*.ini")))
                _file_paths.extend(glob.glob(os.path.join(_config_dir, "*.cfg")))

    _file_paths.sort()
    for _file_path in _file_paths:
        _config_data = read_config_file(config_path=_file_path)
        _config_dict = deep_merge(_config_dict, _config_data)

    return _config_dict


__all__ = [
    "create_dir",
    "remove_dir",
    "remove_dirs",
    "remove_file",
    "remove_files",
    "get_file_checksum",
    "read_yaml_file",
    "read_json_file",
    "read_toml_file",
    "read_ini_file",
    "read_config_file",
    "read_all_configs",
]



from enum import Enum


class WarnEnum(str, Enum):
    ERROR = "ERROR"
    ALWAYS = "ALWAYS"
    DEBUG = "DEBUG"
    IGNORE = "IGNORE"


class TSUnitEnum(str, Enum):
    SECONDS = "SECONDS"
    MILLISECONDS = "MILLISECONDS"
    MICROSECONDS = "MICROSECONDS"
    NANOSECONDS = "NANOSECONDS"


class HashAlgoEnum(str, Enum):
    md5 = "md5"
    sha1 = "sha1"
    sha224 = "sha224"
    sha256 = "sha256"
    sha384 = "sha384"
    sha512 = "sha512"


class ConfigFileFormatEnum(str, Enum):
    YAML = "YAML"
    JSON = "JSON"
    TOML = "TOML"
    INI = "INI"


__all__ = [
    "WarnEnum",
    "TSUnitEnum",
    "HashAlgoEnum",
    "ConfigFileFormatEnum",
]
